import uuid
from textwrap import dedent

from django.db import transaction
from django.utils.functional import cached_property
from drf_spectacular.utils import OpenApiResponse, PolymorphicProxySerializer, extend_schema, extend_schema_view
from rest_framework import permissions, status
from rest_framework.exceptions import PermissionDenied, ValidationError
from rest_framework.generics import CreateAPIView, GenericAPIView, ListCreateAPIView, RetrieveUpdateDestroyAPIView
from rest_framework.response import Response

from arkindex.documents.models import (
    Classification,
    ClassificationState,
    Corpus,
    Element,
    ElementPath,
    MLClass,
    Transcription,
)
from arkindex.documents.serializers.ml import (
    AnnotatedElementSerializer,
    ClassificationCreateSerializer,
    ClassificationMode,
    ClassificationSerializer,
    ClassificationsSelectionSerializer,
    ClassificationsSerializer,
    ElementTranscriptionsBulkSerializer,
    MLClassSerializer,
    TranscriptionBulkSerializer,
    TranscriptionCreateSerializer,
    TranscriptionSerializer,
)
from arkindex.project.filters import SafeSearchFilter
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, SelectionMixin
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.users.models import Role


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateTranscription",
        responses={201: TranscriptionSerializer},
        tags=["transcriptions"],
    )
)
class TranscriptionCreate(ACLMixin, CreateAPIView):
    """
    Create a single transcription attached to an element
    """
    serializer_class = TranscriptionCreateSerializer
    permission_classes = (IsVerified, )

    def get_object(self):
        if not hasattr(self, "element"):
            self.element = super().get_object()
            if not self.has_access(self.element.corpus, Role.Contributor.value):
                raise PermissionDenied(detail="A write access to the element's corpus is required.")
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()

        # We need to use the default database to avoid stale read
        # when creating a transcription on a newly created element.
        # We retrieve the readable objects then check permissions
        # instead of retrieving writable objects directly so as not to
        # get 404_NOT_FOUND errors on elements the user has access to.
        return (
            Element.objects
            .using("default")
            .filter(corpus__in=Corpus.objects.readable(self.request.user))
            .select_related("corpus")
        )

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context["element"] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        obj = serializer.save()
        headers = self.get_success_headers(serializer.data)
        return Response(
            # Use a single transcription serializer for the response
            TranscriptionSerializer(obj).data,
            status=status.HTTP_201_CREATED, headers=headers
        )


@extend_schema(tags=["transcriptions"])
@extend_schema_view(
    get=extend_schema(description="Retrieve a transcription."),
    patch=extend_schema(
        description="Update a transcription's text. "
                    "For non-manual transcriptions, this requires an admin role on the corpus."
    ),
    put=extend_schema(
        description="Update a transcription's text. "
                    "For non-manual transcriptions, this requires an admin role on the corpus."
    ),
    delete=extend_schema(description="Delete a transcription."),
)
class TranscriptionEdit(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a transcription
    Edition and deletion are allowed for manual transcriptions
    Edition and deletion of non-manual transcriptions are allowed for the admin
    """
    serializer_class = TranscriptionSerializer
    permission_classes = (IsVerified, )

    def get_queryset(self):
        if not self.request:
            return Transcription.objects.none()
        return Transcription.objects \
            .select_related("element__corpus") \
            .filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, transcription):
        super().check_object_permissions(request, transcription)
        if self.request.method in permissions.SAFE_METHODS:
            return

        role = Role.Contributor
        detail = "A write access to transcription element corpus is required."
        if transcription.worker_version_id or transcription.worker_run_id:
            role = Role.Admin
            detail = "Only admins can edit non-manual transcription."
        if not self.has_access(transcription.element.corpus, role.value):
            raise PermissionDenied(detail=detail)

        if self.request.method in ("PUT", "PATCH") and transcription.transcription_entities.exists():
            raise PermissionDenied(detail="Transcriptions with entities cannot be modified.")


@extend_schema_view(
    post=extend_schema(
        operation_id="CreateElementTranscriptions",
        tags=["transcriptions"],
        # Return different serializers depending on return_elements
        responses=PolymorphicProxySerializer(
            component_name="ElementTranscriptionsResponse",
            resource_type_field_name="return_elements",
            serializers={
                True: AnnotatedElementSerializer,
                False: ElementTranscriptionsBulkSerializer,
            },
        )
    )
)
class ElementTranscriptionsBulk(CreateAPIView):
    """
    Create multiple sub elements with their transcriptions on a common parent in one transaction
    """
    serializer_class = ElementTranscriptionsBulkSerializer
    permission_classes = (IsVerified, )

    def get_object(self):
        if not hasattr(self, "element"):
            self.element = super().get_object()
            if not self.element.image_id:
                raise ValidationError({"element": ["Cannot create transcriptions on an element without a zone."]})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        # Load elements from the default DB to avoid
        # a stale read issue when an element is created by a worker
        # and this endpoint is immediately used after to create transcriptions
        return Element.objects.filter(
            corpus__in=Corpus.objects.writable(self.request.user)
        ).using("default").select_related("corpus", "image")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context["element"] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        annotations = self.perform_create(serializer)
        headers = self.get_success_headers(serializer.data)

        if serializer.validated_data["return_elements"]:
            # Re-serialize the annotated sub-elements
            data = AnnotatedElementSerializer(annotations, many=True).data
        else:
            data = serializer.data

        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        elt_type = serializer.validated_data["element_type"]
        worker_run = serializer.validated_data["worker_run"]
        annotations = serializer.validated_data["transcriptions"]

        # Retrieve or create elements to attach transcriptions to
        missing_elements = []
        missing_paths = []
        # List existing direct children (with their zones and transcriptions)
        children = {
            element.polygon.wkb: element
            for element in Element
            .objects
            .get_descending(self.element.id)
            .using("default")
            .filter(
                image_id=self.element.image_id,
                type=elt_type,
                paths__path__overlap=[self.element.id],
                paths__path__last=self.element.id
            )
            .only("id", "polygon")
        }
        # Load the paths immediately to avoid iterating over them for each element
        # Use the default DB to avoid stale reads that could cause the new child elements
        # to not be in the parent element's parents
        paths = list(self.element.paths.using("default").values_list("path", flat=True))
        if not paths:
            # Support top level elements, by adding an empty initial path to trigger loops below
            paths = [[]]
        next_path_ordering = self.element.get_next_order()
        for annotation in annotations:
            # Look for a direct children with the right type and zone
            annotation["element"] = children.get(annotation["polygon"].wkb)
            if not annotation["element"]:
                annotation["element"] = Element(
                    id=uuid.uuid4(),
                    image_id=self.element.image_id,
                    polygon=annotation["polygon"],
                    corpus_id=self.element.corpus_id,
                    type=elt_type,
                    name=next_path_ordering + 1,
                    rotation_angle=self.element.rotation_angle,
                    mirrored=self.element.mirrored,
                    worker_version_id=worker_run.version_id,
                    worker_run=worker_run,
                    confidence=annotation.get("element_confidence"),
                )
                # Specify the annotated element has been created
                annotation["created"] = True
                children[annotation["polygon"].wkb] = annotation["element"]
                missing_elements.append(annotation["element"])
                for parent_path in paths:
                    new_path = parent_path + [self.element.id]
                    # Add the children to all of its parent paths
                    missing_paths.append(ElementPath(
                        element=annotation["element"],
                        path=new_path,
                        ordering=next_path_ordering
                    ))
                next_path_ordering += 1
        # Create missing elements and links in bulk
        Element.objects.bulk_create(missing_elements)
        ElementPath.objects.bulk_create(missing_paths)

        # Create transcriptions for the elements if they does not exist
        transcriptions = []
        for annotation in annotations:
            transcription = Transcription(
                element=annotation["element"],
                worker_version_id=worker_run.version_id,
                worker_run=worker_run,
                text=annotation["text"],
                confidence=annotation["confidence"],
                orientation=annotation["orientation"]
            )
            annotation["id"] = transcription.id
            transcriptions.append(transcription)
        # Create missing transcriptions in bulk
        Transcription.objects.bulk_create(transcriptions)

        # Return ordered annotations
        return annotations


@extend_schema_view(post=extend_schema(operation_id="CreateTranscriptions", tags=["transcriptions"]))
class TranscriptionBulk(CreateAPIView):
    """
    Create multiple transcriptions at once on existing elements
    """
    permission_classes = (IsVerified, )
    serializer_class = TranscriptionBulkSerializer


class CorpusMLClassPagination(PageNumberPagination):
    page_size = 40


@extend_schema(tags=["classifications"])
@extend_schema_view(
    get=extend_schema(
        operation_id="ListCorpusMLClasses",
        description=dedent("""
            List available classes in a corpus.

            Requires a **guest** access to the corpus.
        """),
    ),
    post=extend_schema(
        operation_id="CreateMLClass",
        description=dedent("""
            Create an ML class in a corpus.

            Requires an **admin** access to the corpus.
        """),
    )
)
class CorpusMLClassList(CorpusACLMixin, ListCreateAPIView):
    serializer_class = MLClassSerializer
    pagination_class = CorpusMLClassPagination
    # For OpenAPI type discovery: a corpus ID is in the path
    queryset = Corpus.objects.none()
    filter_backends = [SafeSearchFilter]
    search_fields = ["name"]
    permission_classes = (IsVerifiedOrReadOnly, )

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request.method == "POST":
            role = Role.Admin
        return self.get_corpus(self.kwargs["pk"], role=role)

    def check_permissions(self, *args, **kwargs):
        """
        When a user only has a guest access to the corpus and tries to access this endpoint with a browser,
        DRF's HTML renderer will try to get a serializer instance to build an HTML form from it.

        To do so, it goes through the endpoint again with a request that has been faked to look like a POST,
        even though the original request was GET. `get_serializer_context` tries to get the corpus to put in
        the serializer, but `CorpusACLMixin.get_corpus` raises a PermissionDenied because we do not have
        contributor access. This error normally causes a HTTP 403, but here the renderer does not handle it
        and it turns into HTTP 500.

        To avoid this error, the corpus is moved to a cached property that gets evaluated before the renderer
        tries to render the response. That way, a corpus with only guest access is available and no error
        occurs when a fake POST request is used. When an actual POST is made, get_corpus will cause its error
        before the renderer renders a response; the error gets caught and sent to the renderer to show a
        proper HTTP 403.
        """
        super().check_permissions(*args, **kwargs)
        self.corpus

    def get_queryset(self):
        return MLClass.objects.filter(corpus=self.corpus).using("default").order_by("name", "id")

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request.method == "POST":
            context["corpus_id"] = self.corpus
        return context


@extend_schema(tags=["classifications"])
@extend_schema_view(
    get=extend_schema(description=dedent("""
        Retrieve an ML class.

        Requires a **guest** access to the corpus.
    """)),
    patch=extend_schema(description=dedent("""
        Rename an ML class.

        Requires an **admin** access to the corpus.
    """)),
    put=extend_schema(description=dedent("""
        Rename an ML class.

        Requires an **admin** access to the corpus.
    """)),
    delete=extend_schema(description=dedent("""
        Delete an ML class if it is not used by any classification.

        Requires an **admin** access to the corpus.
    """)),
)
class MLClassRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    serializer_class = MLClassSerializer
    queryset = MLClass.objects.none()
    permission_classes = (IsVerified, )
    lookup_url_kwarg = "mlclass"

    @cached_property
    def corpus(self):
        role = Role.Guest
        if self.request and self.request.method != "GET":
            role = Role.Admin

        return self.get_corpus(self.kwargs["corpus"], role=role)

    def get_queryset(self):
        return self.corpus.ml_classes.all()

    def get_serializer_context(self):
        context = super().get_serializer_context()
        context["corpus_id"] = self.corpus
        return context

    def check_object_permissions(self, request, obj):
        if self.request and self.request.method == "DELETE" and obj.classifications.count() != 0:
            raise ValidationError(["This ML class is used by some classifications on this corpus."])


@extend_schema(tags=["classifications"])
class ClassificationCreate(CreateAPIView):
    """
    Add a single classification on an element. Can be used by end users & machine learning tools.
    """
    serializer_class = ClassificationCreateSerializer
    queryset = Classification.objects.all()
    permission_classes = (IsVerified, )

    def perform_create(self, serializer):
        if serializer.validated_data["worker_run"] is None:
            # A manual classification is immediately valid
            serializer.save(
                moderator=self.request.user,
                state=ClassificationState.Validated,
                high_confidence=True,
                confidence=1
            )
        else:
            # A classification produced by a tool is simply saved with its attributes
            serializer.save()


@extend_schema(tags=["classifications"])
class ClassificationBulk(CreateAPIView):
    """
    Create multiple classifications at once on the same element from a single WorkerRun.
    """
    serializer_class = ClassificationsSerializer
    permission_classes = (IsVerified, )


@extend_schema(tags=["classifications"])
class ManageClassificationsSelection(SelectionMixin, CorpusACLMixin, CreateAPIView):
    """
    Manage classifications for a list of selected elements for a specific corpus.
    """
    serializer_class = ClassificationsSelectionSerializer
    permission_classes = (IsVerified, )

    def post(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus = serializer.validated_data["corpus_id"]
        if not self.has_write_access(corpus):
            raise PermissionDenied(detail="You do not have contributor access to this corpus.")

        mode = serializer.validated_data["mode"]
        if mode == ClassificationMode.Create:
            return self.create(corpus, request, *args, **kwargs)
        if mode == ClassificationMode.Validate:
            elements = self.get_selection(corpus.id)
            Classification.objects.filter(
                element__in=elements,
                state=ClassificationState.Pending,
                high_confidence=True
            ).update(state=ClassificationState.Validated)
            return Response(status=status.HTTP_201_CREATED)
        raise NotImplementedError

    def create(self, corpus, request, *args, **kwargs):
        ml_class = MLClass.objects.filter(id=request.data["ml_class"]).first()

        elements = self.get_selection(corpus.id)
        existing_element_ids = set(Classification.objects.filter(
            element_id__in=elements,
            ml_class_id=ml_class.id,
            worker_version_id=None,
        ).values_list("element_id", flat=True))

        classifications = []
        for element in elements.exclude(id__in=existing_element_ids):
            classifications.append(Classification(
                element=element,
                ml_class=ml_class,
                moderator=self.request.user,
                state=ClassificationState.Validated,
                high_confidence=False,
                confidence=1
            ))
        Classification.objects.bulk_create(classifications)

        return Response(
            status=status.HTTP_201_CREATED
        )


@extend_schema(
    tags=["classifications"],
    request=None,
)
class ClassificationModerationActionsMixin(GenericAPIView):
    permission_classes = (IsVerified, )
    serializer_class = ClassificationSerializer

    def get_queryset(self):
        return Classification.objects.filter(element__corpus__in=Corpus.objects.writable(self.request.user))

    def put(self, request, *args, **kwargs):
        instance = self.get_object()
        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=["moderator", "state"])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)


@extend_schema_view(put=extend_schema(operation_id="ValidateClassification"))
class ClassificationValidate(ClassificationModerationActionsMixin):
    """
    Validate an existing classification.
    """
    moderation_action = ClassificationState.Validated


class ClassificationReject(ClassificationModerationActionsMixin):
    """
    Reject an existing classification.

    Manual classifications will be deleted.
    """
    moderation_action = ClassificationState.Rejected

    @extend_schema(
        operation_id="RejectClassification",
        responses={
            200: ClassificationSerializer,
            204: OpenApiResponse(
                response=None,
                description="The classification has been deleted.",
            )
        }
    )
    def put(self, request, *args, **kwargs):
        instance = self.get_object()

        if not instance.worker_version_id:
            # Delete manual classifications upon rejection
            instance.delete()
            return Response(None, status=status.HTTP_204_NO_CONTENT)

        instance.moderator = self.request.user
        instance.state = self.moderation_action
        instance.save(update_fields=["moderator", "state"])
        serializer = self.get_serializer(instance)
        return Response(serializer.data, status=status.HTTP_200_OK)
