from uuid import uuid4

from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus, Element
from arkindex.images.models import ImageServer
from arkindex.process.models import ProcessMode, WorkerRun, WorkerVersion
from arkindex.project.aws import S3FileStatus
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role


class TestCreateElements(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.volume_type = cls.corpus.types.get(slug="volume")
        cls.page_type = cls.corpus.types.get(slug="page")
        cls.act_type = cls.corpus.types.get(slug="act")
        cls.vol = cls.corpus.elements.get(name="Volume 1")
        cls.page = Element.objects.get(name="Volume 1, page 2r")
        cls.image = ImageServer.objects.local.images.create(
            path="kingdom/far/away",
            status=S3FileStatus.Checked,
            width=42,
            height=42,
        )
        cls.huge_image = ImageServer.objects.local.images.create(
            path="kingdom/far/far/away",
            status=S3FileStatus.Checked,
            width=4200,
            height=4200,
        )
        cls.worker_version = WorkerVersion.objects.get(worker__slug="dla")

        cls.worker_run = cls.worker_version.worker_runs.filter(process__mode=ProcessMode.Workers).get()
        cls.custom_version = WorkerVersion.objects.get(worker__slug="custom")
        cls.local_worker_run = WorkerRun.objects.get(version=cls.custom_version, process__creator=cls.user)

        # The image is always the same in API responses
        cls.image_response = {
            "id": str(cls.image.id),
            "s3_url": None,
            "width": 42,
            "height": 42,
            "path": "kingdom/far/away",
            "url": "http://server/kingdom/far/away",
            "status": "checked",
            "server": {
                "display_name": "Test Server",
                "max_height": None,
                "max_width": None,
                "url": "http://server",
            }
        }

    def make_create_request(self, name="default", corpus=None, elt_type="volume", **options):
        request = {
            "path": reverse("api:elements-create"),
            "data": {
                "type": elt_type,
                "name": name,
                "corpus": corpus or str(self.corpus.id),
            },
            "format": "json",
        }
        # Optional parameters
        request["data"].update(options)
        return request

    def test_not_verified(self):
        self.assertEqual(
            self.client.post(**self.make_create_request("no")).status_code,
            status.HTTP_403_FORBIDDEN
        )

    def test_volume(self):
        # Create a Volume
        self.client.force_login(self.user)
        request = self.make_create_request("my new volume")
        with self.assertNumQueries(7):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        self.assertIn("id", data)
        volume = Element.objects.get(id=data["id"])
        self.assertDictEqual(
            data,
            {
                "id": str(volume.id),
                "type": volume.type.slug,
                "name": volume.name,
                "thumbnail_url": volume.thumbnail.s3_url,
                "thumbnail_put_url": volume.thumbnail.s3_put_url,
                "confidence": None,
                "creator": "Test user",
                "rights": ["read", "write", "admin"],
                "classifications": [],
                "corpus": {
                    "id": str(volume.corpus.id),
                    "name": volume.corpus.name,
                    "public": volume.corpus.public
                },
                "zone": None,
                "rotation_angle": 0,
                "mirrored": False,
                "created": volume.created.isoformat().replace("+00:00", "Z"),
                "worker_run": None
            }
        )
        self.assertEqual(volume.name, "my new volume")
        self.assertEqual(volume.type, self.volume_type)
        self.assertEqual(volume.creator, self.user)
        self.assertListEqual(volume.paths.get().path, [])

    def test_volume_page(self):
        # Add a page with an Image to an existing volume
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams",
            image=str(self.image.id),
        )
        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data["id"])
        self.assertDictEqual(
            data,
            {
                "id": str(page.id),
                "name": page.name,
                "type": page.type.slug,
                "thumbnail_put_url": None,
                "thumbnail_url": None,
                "created": page.created.isoformat().replace("+00:00", "Z"),
                "creator": "Test user",
                "rights": ["read", "write", "admin"],
                "confidence": None,
                "classifications": [],
                "corpus": {
                    "id": str(page.corpus.id),
                    "name": page.corpus.name,
                    "public": page.corpus.public
                },
                "zone": {
                    "id": str(page.id),
                    "image": self.image_response,
                    "polygon": [[0, 0], [0, 42], [42, 42], [42, 0], [0, 0]],
                    "url": page.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "worker_run": None
            }
        )
        self.assertEqual(page.name, "The castle of my dreams")
        self.assertEqual(page.type, self.page_type)
        self.assertEqual(page.creator, self.user)
        self.assertListEqual(page.paths.get().path, [self.vol.id])

    def test_inherit_parent_image(self):
        """
        Creating an element with a parent, a polygon and no image should inherit the parent's image
        """
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.vol.image = self.image
        self.vol.polygon = polygon
        self.vol.save()
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams",
            polygon=polygon,
        )
        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data["id"])
        self.assertEqual(response.json()["zone"], {
            "id": str(page.id),
            "image": self.image_response,
            "polygon": polygon,
            "url": page.iiif_url,
        })

    def test_volume_act(self):
        # Create an volume child of type Act
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            name="Castle story",
            elt_type="act"
        )
        with self.assertNumQueries(11):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        act = Element.objects.get(id=response.json()["id"])
        self.assertEqual(act.name, "Castle story")
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.creator, self.user)
        self.assertListEqual(act.paths.get().path, [self.vol.id])

    def test_polygon_no_image(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            elt_type="page",
            name="The castle of my dreams again",
            polygon=[[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]],
        )
        with self.assertNumQueries(4):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "polygon": ["An image or a parent with an image is required to create an element with a polygon."],
        })

    def test_polygon_negative(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams again",
            image=str(self.image.id),
            polygon=[[0, 0], [10, 10], [40, 40], [-10, 10], [0, 0]]
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "polygon": {"3": {"0": ["Ensure this value is greater than or equal to 0."]}}
        })

    def test_polygon(self):
        # Create an element with a polygon to an existing volume
        polygon = [[10, 10], [10, 40], [40, 40], [40, 10], [10, 10]]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams again",
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        page = self.corpus.elements.get(id=data["id"])
        self.assertDictEqual(
            data,
            {
                "id": str(page.id),
                "name": page.name,
                "type": page.type.slug,
                "thumbnail_put_url": None,
                "thumbnail_url": None,
                "confidence": None,
                "creator": "Test user",
                "rights": ["read", "write", "admin"],
                "classifications": [],
                "corpus": {
                    "id": str(page.corpus.id),
                    "name": page.corpus.name,
                    "public": page.corpus.public
                },
                "zone": {
                    "id": str(page.id),
                    "polygon": polygon,
                    "image": self.image_response,
                    "url": page.iiif_url,
                },
                "rotation_angle": 0,
                "mirrored": False,
                "created": page.created.isoformat().replace("+00:00", "Z"),
                "worker_run": None
            }
        )
        self.assertEqual(page.name, "The castle of my dreams again")
        self.assertEqual(page.type, self.page_type)

    def test_huge_polygon(self):
        # Create a polygon with points in a reversed order, because this is what ST_Normalize would return
        # (this makes assertions easier than handling a reordering)
        polygon = [[0, 0]] + [[i, i] for i in range(3000, -1, -1)]
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams again",
            image=str(self.huge_image.id),
            polygon=polygon
        )
        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED, response.json())
        self.assertListEqual(response.json()["zone"]["polygon"], polygon)

    def test_parent_different_corpus(self):
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name="new")
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_corpus.types.create(slug="act")
        request = self.make_create_request(
            parent=str(self.vol.id),
            corpus=str(new_corpus.id),
            elt_type="act"
        )
        with self.assertNumQueries(5):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "corpus": ["Parent and child must be in the same corpus"]
        })

    def test_filter_type(self):
        """
        Ensure types are filtered by corpus
        """
        self.assertTrue(self.corpus.types.filter(slug="volume").exists())
        self.client.force_login(self.user)
        new_corpus = Corpus.objects.create(name="new")
        new_corpus.memberships.create(user=self.user, level=Role.Contributor.value)
        new_vol_type = new_corpus.types.create(slug="volume", display_name="Volume", folder=True)
        request = self.make_create_request(
            corpus=str(new_corpus.id),
            elt_type="volume",
            name="something",
        )
        with self.assertNumQueries(7):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        new_vol = new_corpus.elements.get()
        self.assertEqual(new_vol.type, new_vol_type)
        self.assertEqual(new_vol.type.corpus, new_corpus)

    def test_wrong_type(self):
        self.assertFalse(self.corpus.types.filter(slug="kartoffelsalad").exists())
        self.client.force_login(self.user)
        request = self.make_create_request(
            corpus=str(self.corpus.id),
            elt_type="kartoffelsalad",
            name="something",
        )
        with self.assertNumQueries(4):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "type": ["Type 'kartoffelsalad' not found in corpus Unit Tests"],
        })

    def test_image_not_checked(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            image=str(self.image.id)
        )
        self.image.status = S3FileStatus.Error
        self.image.save()
        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()["id"])
        self.assertEqual(element.image.status, S3FileStatus.Error)

    def test_image_zero_dimensions(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            image=str(self.image.id)
        )
        self.image.width = 0
        self.image.height = 0
        self.image.status = S3FileStatus.Unchecked
        self.image.save()
        with self.assertNumQueries(6):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {"image": ["This image does not have valid dimensions."]})

    def test_slim_output_simple(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name="slim output !",
        )
        request["path"] += "?slim_output=true"
        with self.assertNumQueries(6):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()["id"])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_slim_output(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            name="slim output !",
            parent=str(self.vol.id),
            elt_type="page",
            image=str(self.image.id),
        )
        request["path"] += "?slim_output=true"
        with self.assertNumQueries(11):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        element = self.corpus.elements.get(id=response.json()["id"])
        self.assertDictEqual(response.json(), {
            "id": str(element.id),
        })

    def test_rotation_mirroring(self):
        self.client.force_login(self.user)

        self.vol.rotation_angle = 180
        self.vol.mirrored = True
        self.vol.save()

        # Rotation and mirroring sent in the request, and returned after inheritance from the volume
        parameters = [
            (0, False, 0, False),
            (None, False, 180, False),
            (None, None, 180, True),
            (0, None, 0, True),
        ]

        for rotation_angle, mirrored, expected_rotation_angle, expected_mirrored in parameters:
            with self.subTest(rotation_angle=rotation_angle, mirrored=mirrored):
                request = self.make_create_request(
                    parent=str(self.vol.id),
                    elt_type="page",
                    name="The castle of my dreams again",
                    image=str(self.image.id),
                )

                if rotation_angle is not None:
                    request["data"]["rotation_angle"] = rotation_angle
                if mirrored is not None:
                    request["data"]["mirrored"] = mirrored

                with self.assertNumQueries(12):
                    response = self.client.post(**request)
                    self.assertEqual(response.status_code, status.HTTP_201_CREATED)

                data = response.json()
                self.assertEqual(data["rotation_angle"], expected_rotation_angle)
                self.assertEqual(data["mirrored"], expected_mirrored)

    def test_rotation_validation(self):
        self.client.force_login(self.user)

        parameters = [
            (-1, "Ensure this value is greater than or equal to 0."),
            (360, "Ensure this value is less than or equal to 359."),
            (1.3, "A valid integer is required."),
        ]

        for rotation_angle, expected_message in parameters:
            with self.subTest(rotation_angle=rotation_angle):
                request = self.make_create_request(
                    parent=str(self.vol.id),
                    elt_type="page",
                    name="The castle of my dreams again",
                    image=str(self.image.id),
                    rotation_angle=rotation_angle,
                )

                with self.assertNumQueries(6):
                    response = self.client.post(**request)
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "rotation_angle": [expected_message]
                })

    def test_outside_image(self):
        """
        Cannot create an elements outside of its image
        """
        self.client.force_login(self.user)
        polygon = [[32, 32], [32, 56], [56, 56], [56, 32], [32, 32]]
        request = self.make_create_request(
            parent=str(self.page.id),
            elt_type="act",
            name="Out Of Bounds",
            image=str(self.image.id),
            polygon=polygon
        )
        with self.assertNumQueries(6):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertDictEqual(response.json(), {
            "polygon": ["An element's polygon must not exceed its image's bounds."]
        })

    def test_confidence(self):
        self.client.force_login(self.user)
        request = self.make_create_request(
            parent=str(self.vol.id),
            elt_type="page",
            name="The castle of my dreams again",
            image=str(self.image.id),
            confidence=0.42,
        )

        with self.assertNumQueries(12):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        data = response.json()
        element = self.corpus.elements.get(id=data["id"])
        self.assertEqual(element.confidence, 0.42)

    def test_confidence_validation(self):
        self.client.force_login(self.user)

        parameters = [
            (-1, "Ensure this value is greater than or equal to 0."),
            (2, "Ensure this value is less than or equal to 1."),
            ("Very", "A valid number is required."),
        ]

        for confidence, expected_message in parameters:
            with self.subTest(confidence=confidence):
                request = self.make_create_request(
                    parent=str(self.vol.id),
                    elt_type="page",
                    name="The castle of my dreams again",
                    image=str(self.image.id),
                    confidence=confidence,
                )

                with self.assertNumQueries(6):
                    response = self.client.post(**request)
                    self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

                self.assertDictEqual(response.json(), {
                    "confidence": [expected_message]
                })

    def test_worker_run_local(self):
        """
        A regular user can create an element with a WorkerRun of their own local process
        """
        self.client.force_login(self.local_worker_run.process.creator)

        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(self.local_worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        act = Element.objects.get(id=response.json()["id"])
        self.assertEqual(act.name, "Castle story")
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.worker_run, self.local_worker_run)
        self.assertEqual(act.worker_version, self.local_worker_run.version)
        self.assertIsNone(act.creator)
        self.assertDictEqual(response.json()["worker_run"], {
            "id": str(self.local_worker_run.id),
            "summary": "Worker Custom worker @ version 1",
        })

    def test_worker_run_task(self):
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(self.worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        act = Element.objects.get(id=response.json()["id"])
        self.assertEqual(act.name, "Castle story")
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.worker_run, self.worker_run)
        self.assertEqual(act.worker_version, self.worker_version)
        self.assertIsNone(act.creator)
        self.assertDictEqual(response.json()["worker_run"], {
            "id": str(self.worker_run.id),
            "summary": "Worker Document layout analyser @ version 1",
        })

    def test_worker_run_local_task_auth(self):
        """
        An element can be created with a WorkerRun of a Local process
        even when authenticated as a Ponos task from a different process
        """
        # Take the user's local process, so that the user's rights will take over the Ponos task auth
        local_process = self.worker_run.process.creator.processes.get(mode=ProcessMode.Local)
        local_worker_run = local_process.worker_runs.get()
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(local_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        act = Element.objects.get(id=response.json()["id"])
        self.assertEqual(act.name, "Castle story")
        self.assertEqual(act.type, self.act_type)
        self.assertEqual(act.worker_run, self.local_worker_run)
        self.assertEqual(act.worker_version, self.local_worker_run.version)
        self.assertIsNone(act.creator)
        self.assertDictEqual(response.json()["worker_run"], {
            "id": str(self.local_worker_run.id),
            "summary": "Worker Custom worker @ version 1",
        })

    def test_worker_run_non_local(self):
        """
        A regular user cannot create an element with a WorkerRun of a non-local process
        """
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(self.worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_worker_run_other_user(self):
        """
        A regular user cannot create an element with a WorkerRun of someone else's local process
        """
        worker_run = self.user.processes.get(mode=ProcessMode.Local).worker_runs.first()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(worker_run.id),
                },
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Ponos task authentication is required to use a WorkerRun of a process other than the user's local process."],
        })

    def test_worker_run_other_process(self):
        """
        A Ponos task cannot create an element with a WorkerRun of another process
        """
        # Create a process on the same user, to check that being the process creator has no effect
        process2 = self.worker_run.process.creator.processes.create(
            mode=ProcessMode.Workers,
            corpus=self.corpus,
        )
        other_worker_run = process2.worker_runs.create(version=self.worker_run.version, parents=[])
        self.worker_run.process.run()
        task = self.worker_run.process.tasks.first()

        with self.assertNumQueries(4):
            response = self.client.post(
                reverse("api:elements-create"),
                {
                    "corpus": str(self.corpus.id),
                    "name": "Castle story",
                    "type": "act",
                    "worker_run_id": str(other_worker_run.id),
                },
                HTTP_AUTHORIZATION=f"Ponos {task.token}",
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {
            "worker_run_id": ["Only the WorkerRuns of the authenticated task's process may be used."],
        })

    def test_unknown_worker_run(self):
        self.client.force_login(self.user)
        random_uuid = str(uuid4())
        request = self.make_create_request(
            name="Castle story",
            elt_type="act",
            worker_run_id=random_uuid,
        )
        with self.assertNumQueries(5):
            response = self.client.post(**request)
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)

        self.assertDictEqual(response.json(), {"worker_run_id": [f'Invalid pk "{random_uuid}" - object does not exist.']})
