import copy
import uuid
from datetime import datetime, timedelta, timezone
from io import BytesIO
from itertools import combinations
from unittest.mock import call, patch, seal

from botocore.exceptions import ClientError
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.documents.models import Corpus
from arkindex.ponos.models import Agent, AgentMode, Farm, State, Task
from arkindex.process.models import Process, ProcessMode, Revision, WorkerVersion
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Right, Role, User


@override_settings(PONOS_LOG_TAIL=42)
class TestAPI(FixtureAPITestCase):

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        cls.page_1 = cls.corpus.elements.get(name="Volume 1, page 1r")
        cls.page_2 = cls.corpus.elements.get(name="Volume 1, page 2r")
        cls.rev = Revision.objects.first()
        cls.process = Process.objects.get(mode=ProcessMode.Workers)
        cls.process.run()
        cls.task1, cls.task2, cls.task3 = cls.process.tasks.all().order_by("depth")

        # Brand new user and corpus with no preexisting rights
        new_user = User.objects.create(email="another@user.com")
        new_corpus = Corpus.objects.create(name="corpus the second")
        Right.objects.create(user=new_user, content_object=new_corpus, level=50)
        cls.process2 = Process.objects.create(
            mode=ProcessMode.Workers,
            creator=new_user,
            corpus=new_corpus,
        )
        cls.process2.run()
        cls.task4 = cls.process2.tasks.first()

        cls.dla = WorkerVersion.objects.get(worker__slug="dla")
        cls.recognizer = WorkerVersion.objects.get(worker__slug="reco")

        cls.farm = Farm.objects.first()
        cls.docker_agent = Agent.objects.create(
            mode=AgentMode.Docker,
            farm=cls.farm,
            last_ping=datetime.now(timezone.utc),
            cpu_cores=42,
            cpu_frequency=42e8,
            ram_total=42e3
        )
        cls.slurm_agent = Agent.objects.create(
            mode=AgentMode.Slurm,
            farm=cls.farm,
            last_ping=datetime.now(timezone.utc),
        )

    @property
    def docker_task_transitions(self):
        return (
            (State.Unscheduled, State.Pending),
            (State.Pending, State.Error),
            (State.Pending, State.Running),
            (State.Running, State.Completed),
            (State.Running, State.Failed),
            (State.Running, State.Error),
            (State.Stopping, State.Stopped),
            (State.Stopping, State.Error),
        )

    @property
    def slurm_task_transitions(self):
        return self.docker_task_transitions + (
            (State.Pending, State.Completed),
            (State.Pending, State.Failed),
        )

    def test_task_details_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.aws.s3")
    def test_task_details_own_task(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        with self.assertNumQueries(3):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task1.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task1.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "original_task_id": None,
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
                "requires_gpu": False,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_task_details_other_task_no_access_forbidden(self):
        # Make corpus private just for this test
        self.corpus.public = False
        self.corpus.save()

        with self.assertNumQueries(2):
            resp = self.client.get(
                reverse("api:task-details", args=[self.task2.id]),
                HTTP_AUTHORIZATION=f"Ponos {self.task4.token}",
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_task_details_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.mixins.get_max_level")
    def test_task_details_requires_process_guest(self, get_max_level_mock):
        get_max_level_mock.return_value = None
        self.client.force_login(self.user)

        with self.assertNumQueries(3):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

        self.assertEqual(get_max_level_mock.call_count, 1)
        self.assertEqual(get_max_level_mock.call_args, call(self.user, self.corpus))

    @patch("arkindex.project.aws.s3")
    def test_task_details_process_level_corpus(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.process.creator = self.superuser
        self.process.save()
        self.client.force_login(self.user)

        for role in Role:
            s3_mock.reset_mock()
            s3_mock.reset_mock()
            # Recreate the BytesIO each time, because its contents get consumed each time the API is called
            s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}

            with self.subTest(role=role):
                self.corpus.memberships.filter(user=self.user).update(level=role.value)

                with self.assertNumQueries(4):
                    resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                    self.assertEqual(resp.status_code, status.HTTP_200_OK)

                self.assertDictEqual(
                    resp.json(),
                    {
                        "id": str(self.task1.id),
                        "run": 0,
                        "depth": 0,
                        "slug": "initialisation",
                        "state": "unscheduled",
                        "parents": [],
                        "original_task_id": None,
                        "logs": "Failed successfully",
                        "full_log": "http://somewhere",
                        "extra_files": {},
                        "agent": None,
                        "gpu": None,
                        "shm_size": None,
                        "requires_gpu": False,
                    },
                )

                self.assertEqual(s3_mock.Object.call_count, 2)
                self.assertEqual(s3_mock.Object().get.call_count, 1)
                self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
                self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
                self.assertEqual(
                    s3_mock.meta.client.generate_presigned_url.call_args,
                    call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
                )

    @patch("arkindex.project.aws.s3")
    def test_task_details(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "unscheduled",
                "parents": [],
                "original_task_id": None,
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
                "requires_gpu": False,
            },
        )

        self.assertEqual(s3_mock.Object.call_count, 2)
        self.assertEqual(s3_mock.Object().get.call_count, 1)
        self.assertEqual(s3_mock.Object().get.call_args, call(Range="bytes=-42"))
        self.assertEqual(s3_mock.meta.client.generate_presigned_url.call_count, 1)
        self.assertEqual(
            s3_mock.meta.client.generate_presigned_url.call_args,
            call("get_object", Params={"Bucket": "ponos", "Key": "somelog"}),
        )

    def test_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level")
    def test_update_task_requires_process_admin_corpus(self, get_max_level_mock):
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                get_max_level_mock.return_value = role.value if role else None

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.project.aws.s3")
    def test_update_running_task_state_stopping(self, s3_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        self.task1.state = State.Running
        self.task1.save()
        self.client.force_login(self.superuser)

        with self.assertNumQueries(5):
            resp = self.client.put(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_200_OK)

        self.assertDictEqual(
            resp.json(),
            {
                "id": str(self.task1.id),
                "run": 0,
                "depth": 0,
                "slug": "initialisation",
                "state": "stopping",
                "parents": [],
                "original_task_id": None,
                "logs": "Failed successfully",
                "full_log": "http://somewhere",
                "extra_files": {},
                "agent": None,
                "gpu": None,
                "shm_size": None,
                "requires_gpu": False,
            },
        )
        self.task1.refresh_from_db()
        self.assertEqual(self.task1.state, State.Stopping)

    def test_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        states.remove(State.Stopping)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]},
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_update_running_task_non_pending(self):
        states = list(State)
        states.remove(State.Stopping)
        self.task1.state = State.Running
        self.task1.save()

        self.client.force_login(self.superuser)
        for state in states:
            with self.subTest(state=state):
                with self.assertNumQueries(3):
                    resp = self.client.put(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": state.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state Running to state {state} is forbidden."]},
                )

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_partial_update_task_from_docker_agent_allowed_transitions(self, notify_mock, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in self.docker_task_transitions:
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_to)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.serializers.notify_process_completion")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_partial_update_task_from_slurm_agent_allowed_transitions(self, notify_mock, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Authenticate from a possible agent
        custom_user = copy.copy(self.slurm_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in self.slurm_task_transitions:
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_to)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_partial_update_task_from_docker_agent_forbidden_transitions(self, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Mock authenticating from an agent
        custom_user = copy.copy(self.docker_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in combinations(State, 2):
            if (state_from, state_to) in self.docker_task_transitions:
                continue
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(resp.json(), {
                    "state": [f"Transition from state {state_from} to state {state_to} is forbidden."]
                })
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_from)

    @patch("arkindex.users.models.User.objects.get")
    @patch("arkindex.project.aws.s3")
    @patch("arkindex.ponos.api.TaskDetails.permission_classes", tuple())
    def test_partial_update_task_from_slurm_agent_forbidden_transitions(self, s3_mock, get_user_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {"Body": BytesIO(b"Failed successfully")}
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        # Mock authenticating from an agent
        custom_user = copy.copy(self.slurm_agent)
        custom_user.is_active = True
        custom_user.is_authenticated = True

        get_user_mock.return_value = custom_user
        self.client.force_login(self.user)

        for (state_from, state_to) in combinations(State, 2):
            if (state_from, state_to) in self.slurm_task_transitions:
                continue
            with self.subTest(state_from=state_from, state_to=state_to):
                self.task1.state = state_from
                self.task1.save()
                resp = self.client.patch(
                    reverse("api:task-details", args=[self.task1.id]),
                    data={"state": state_to.value},
                )
                self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(resp.json(), {
                    "state": [f"Transition from state {state_from} to state {state_to} is forbidden."]
                })
                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state_from)


    def test_partial_update_task_requires_login(self):
        with self.assertNumQueries(0):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    def test_partial_update_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)

        with self.assertNumQueries(2):
            resp = self.client.patch(
                reverse("api:task-details", args=[self.task1.id]),
                data={"state": State.Stopping.value},
            )
            self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN)

    @patch("arkindex.users.utils.get_max_level")
    def test_partial_update_task_requires_process_admin_corpus(self, get_max_level_mock):
        self.corpus.public = False
        self.corpus.save()
        self.client.force_login(self.user)

        for role in [None, Role.Guest, Role.Contributor]:
            with self.subTest(role=role):
                get_max_level_mock.return_value = role.value if role else None

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_403_FORBIDDEN, resp.json())

    def test_partial_update_non_running_task_state_stopping(self):
        states = list(State)
        states.remove(State.Running)
        states.remove(State.Stopping)
        self.client.force_login(self.superuser)

        for state in states:
            with self.subTest(state=state):
                self.task1.state = state
                self.task1.save()

                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": State.Stopping.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state {state} to state Stopping is forbidden."]},
                )

                self.task1.refresh_from_db()
                self.assertEqual(self.task1.state, state)

    def test_partial_update_running_task_non_pending(self):
        states = list(State)
        states.remove(State.Stopping)
        self.task1.state = State.Running
        self.task1.save()

        self.client.force_login(self.superuser)
        for state in states:
            with self.subTest(state=state):
                with self.assertNumQueries(3):
                    resp = self.client.patch(
                        reverse("api:task-details", args=[self.task1.id]),
                        data={"state": state.value},
                    )
                    self.assertEqual(resp.status_code, status.HTTP_400_BAD_REQUEST)
                self.assertDictEqual(
                    resp.json(),
                    {"state": [f"Transition from state Running to state {state} is forbidden."]},
                )

    @patch("arkindex.project.aws.s3")
    def test_task_logs_unicode_error(self, s3_mock):
        """
        Ensure the TaskLogs.latest property is able to handle sliced off Unicode characters.
        Since we fetch the latest logs from S3 using `Range: bytes=-N`, sometimes the characters
        can have a missing byte and cause a UnicodeDecodeError, such as U+00A0 (non-breaking space)
        or U+00A9 (copyright symbol).
        """
        with self.assertRaises(UnicodeDecodeError):
            b"\xa0Failed successfully".decode("utf-8")

        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"\xa0Failed successfully")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        self.client.force_login(self.superuser)
        with self.assertNumQueries(4):
            resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
            self.assertEqual(resp.status_code, status.HTTP_200_OK)
        data = resp.json()

        self.assertEqual(data["logs"], "�Failed successfully")

    @patch("arkindex.project.aws.s3")
    def test_task_logs_errors(self, s3_mock):
        """
        Ensure the TaskLogs.latest property handles missing log files and returns an empty string,
        or handles an InvalidRange error returned only by Ceph when the log file exists but has zero bytes.
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.side_effect = ClientError({"Error": {"Code": ""}}, "get_object")
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        self.client.force_login(self.superuser)

        for code in ["NoSuchKey", "InvalidRange"]:
            with self.subTest(code=code), self.assertNumQueries(4):
                s3_mock.Object.return_value.get.side_effect.response["Error"]["Code"] = str(code)

                resp = self.client.get(reverse("api:task-details", args=[self.task1.id]))
                self.assertEqual(resp.status_code, status.HTTP_200_OK)
                self.assertEqual(resp.json()["logs"], "")

    def test_restart_task_requires_login(self):
        with self.assertNumQueries(0):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid.uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_restart_task_requires_verified(self):
        self.user.verified_email = False
        self.user.save()
        self.client.force_login(self.user)
        with self.assertNumQueries(2):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid.uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    def test_restart_task_not_found(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(6):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(uuid.uuid4())})
            )
            self.assertEqual(response.status_code, status.HTTP_404_NOT_FOUND)

    @patch("arkindex.project.mixins.get_max_level")
    def test_restart_task_forbidden(self, get_max_level_mock):
        """An admin access to the process is required"""
        get_max_level_mock.return_value = Role.Guest.value
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["You do not have an admin access to the process of this task."],
            )

    def test_restart_task_non_final_state(self):
        self.client.force_login(self.user)
        with self.assertNumQueries(7):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["Task's state must be in a final state to be restarted."],
            )

    def test_restart_task_already_restarted(self):
        self.client.force_login(self.user)
        self.task1.slug = self.task1.slug + "_old1"
        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.original_task_id = self.task1.id
        self.task2.save()
        with self.assertNumQueries(8):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
            self.assertListEqual(
                response.json(),
                ["This task has already been restarted."],
            )

    @override_settings(PONOS_RQ_EXECUTION=False)
    @patch("arkindex.project.aws.s3")
    def test_restart_task(self, s3_mock):
        """
        From:
               task2_old1
             ↗
        task1 → task2 → task3
                   ↘      ↗
                    task 4

        To:
               task2_old1
             ↗
        task1  → task2_old2
             ↘
               task2 → task3
                  ↘      ↗
                   task 4
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        task4 = self.process.tasks.create(run=self.task1.run, depth=1)
        task4.parents.add(self.task2)
        task4.children.add(self.task3)
        task_2_slug = self.task2.slug

        with patch("django.utils.timezone.now") as mock_now:
            mock_now.return_value = datetime.now(timezone.utc) + timedelta(minutes=1)
            old_task_2 = self.process.tasks.create(run=self.task1.run, depth=1, slug=f"{task_2_slug}_old1")
            old_task_2.state = State.Error.value
            old_task_2.original_task_id = self.task1.id
            old_task_2.save()

        old_task_2.parents.add(self.task1)

        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.state = State.Error.value
        self.task2.requires_gpu = True
        self.task2.agent = self.docker_agent
        self.task2.gpu = self.docker_agent.gpus.create(
            id=uuid.uuid4(),
            name="gee pee you",
            index=0,
            ram_total=42e9,
        )
        self.task2.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            with patch("django.utils.timezone.now") as mock_now:
                mock_now.return_value = datetime.now(timezone.utc) + timedelta(minutes=2)
                response = self.client.post(
                    reverse("api:task-restart", kwargs={"pk": str(self.task2.id)})
                )
                self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.process.tasks.count(), 6)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 1,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task2.id),
                "parents": [str(self.task1.id)],
                "run": 0,
                "shm_size": None,
                "slug": task_2_slug,
                "state": "pending",
                "requires_gpu": True,
            },
        )
        self.assertQuerySetEqual(self.task2.children.all(), Task.objects.none())
        self.task2.refresh_from_db()
        self.assertEqual(self.task2.slug, f"{task_2_slug}_old2")
        self.assertNotEqual(self.task2.token, restarted_task.token)
        self.assertIsNone(restarted_task.agent_id)
        self.assertIsNone(restarted_task.gpu_id)
        self.assertQuerySetEqual(
            restarted_task.children.all(),
            Task.objects.filter(id__in=[self.task3.id, task4.id]),
        )

    @override_settings(PONOS_RQ_EXECUTION=False)
    @patch("arkindex.project.aws.s3")
    def test_restart_task_no_previous_restart(self, s3_mock):
        """
        From:
        task1 → task2 → task3
                   ↘      ↗
                    task 4

        To:
        task1  → task2_old1
             ↘
               task2 → task3
                  ↘      ↗
                   task 4
        """
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"

        task4 = self.process.tasks.create(run=self.task1.run, depth=1)
        task4.parents.add(self.task2)
        task4.children.add(self.task3)
        task_2_slug = self.task2.slug
        self.task1.state = State.Completed.value
        self.task1.save()
        self.task2.state = State.Error.value
        self.task2.save()

        self.client.force_login(self.user)
        with self.assertNumQueries(14):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task2.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        self.assertEqual(self.process.tasks.count(), 5)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 1,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task2.id),
                "parents": [str(self.task1.id)],
                "run": 0,
                "shm_size": None,
                "slug": task_2_slug,
                "state": "pending",
                "requires_gpu": False,
            },
        )
        self.assertQuerySetEqual(self.task2.children.all(), Task.objects.none())
        self.task2.refresh_from_db()
        self.assertEqual(self.task2.slug, f"{task_2_slug}_old1")
        self.assertQuerySetEqual(
            restarted_task.children.all(),
            Task.objects.filter(id__in=[self.task3.id, task4.id]),
        )

    @override_settings(PONOS_RQ_EXECUTION=True)
    @patch("arkindex.ponos.tasks.run_task_rq.delay")
    @patch("arkindex.project.aws.s3")
    def test_restart_task_rq(self, s3_mock, delay_mock):
        s3_mock.Object.return_value.bucket_name = "ponos"
        s3_mock.Object.return_value.key = "somelog"
        s3_mock.Object.return_value.get.return_value = {
            "Body": BytesIO(b"Task has been restarted")
        }
        s3_mock.meta.client.generate_presigned_url.return_value = "http://somewhere"
        seal(s3_mock)

        delay_mock.return_value = None
        seal(delay_mock)

        self.task1.state = State.Failed.value
        self.task1.save()

        self.client.force_login(self.user)

        with self.assertNumQueries(13), self.captureOnCommitCallbacks(execute=True):
            response = self.client.post(
                reverse("api:task-restart", kwargs={"pk": str(self.task1.id)})
            )
            self.assertEqual(response.status_code, status.HTTP_201_CREATED)

        self.assertEqual(self.process.tasks.count(), 4)
        restarted_task = self.process.tasks.latest("created")
        self.assertDictEqual(
            response.json(),
            {
                "id": str(restarted_task.id),
                "depth": 0,
                "agent": None,
                "extra_files": {},
                "full_log": "http://somewhere",
                "gpu": None,
                "logs": "Task has been restarted",
                "original_task_id": str(self.task1.id),
                "parents": [],
                "run": 0,
                "shm_size": None,
                "slug": self.task1.slug,
                "state": "pending",
                "requires_gpu": False,
            },
        )

        self.assertEqual(delay_mock.call_count, 1)
        self.assertEqual(delay_mock.call_args, call(restarted_task))
